#!/usr/bin/env python3

import http.client
import io
import json
import logging
import sys
import threading
import urllib.parse

from typing import Any
from typing import Dict
from typing import Optional

import fastavro
import pkg_resources
import singer

from jsonschema import FormatChecker
from jsonschema.validators import Draft4Validator
from smart_open import open

from target_avro.converter import json2avro
from target_avro.converter import str2datetime


REQUIRED_CONFIG_KEYS = ["prefix"]

logger = singer.get_logger()


def emit_state(state: Optional[Dict[str, Any]]) -> None:
    if state is not None:
        line = json.dumps(state)
        logger.debug("Emitting state {}".format(line))
        sys.stdout.write("{}\n".format(line))
        sys.stdout.flush()


def persist_lines(
    config: Dict[str, Any], lines: io.TextIOWrapper
) -> Optional[Dict[str, Any]]:
    state = None
    objects: Dict[str, Any] = {}
    schemas: Dict[str, Dict[str, Any]] = {}
    writers: Dict[str, fastavro.write.Writer] = {}
    key_properties = {}
    validators: Dict[str, Draft4Validator] = {}

    # Loop over lines from stdin
    for line in lines:
        try:
            o = json.loads(line)
        except json.decoder.JSONDecodeError:
            logger.error("Unable to parse:\n{}".format(line))
            raise

        if "type" not in o:
            raise Exception("Line is missing required key 'type': {}".format(line))
        t = o["type"]

        if t == "RECORD":
            if "stream" not in o:
                raise Exception(
                    "Line is missing required key 'stream': {}".format(line)
                )
            if o["stream"] not in schemas:
                raise Exception(
                    "A record for stream {} was encountered before a corresponding schema".format(
                        o["stream"]
                    )
                )
            record = o["record"]
            validators[o["stream"]].validate(record)
            str2datetime(record, schemas[o["stream"]]["properties"])
            writers[o["stream"]].write(record)
            state = None
        elif t == "STATE":
            logger.debug("Setting state to {}".format(o["value"]))
            state = o["value"]
        elif t == "SCHEMA":
            if "stream" not in o:
                raise Exception(
                    "Line is missing required key 'stream': {}".format(line)
                )
            stream = o["stream"]
            schemas[stream] = o["schema"]
            validators[stream] = Draft4Validator(
                o["schema"], format_checker=FormatChecker()
            )
            if "key_properties" not in o:
                raise Exception("key_properties field is required")
            key_properties[stream] = o["key_properties"]

            props = o["schema"]["properties"]
            logger.debug(" ".join([t, stream, "JSON", json.dumps(props)]))
            fields = json2avro(props, o["schema"].get("required", []))
            logger.debug(" ".join([t, stream, "Avro", json.dumps(fields)]))

            f = open("{}.{}.avro".format(config["prefix"], stream), "wb")
            objects[stream] = f
            writers[stream] = fastavro.write.Writer(
                f,
                fastavro.parse_schema(
                    {
                        "fields": fields,
                        "name": stream,
                        "namespace": "app.kageboushi.integrations.v1",
                        "type": "record",
                    }
                ),
            )
        else:
            raise Exception(
                "Unknown message type {} in message {}".format(o["type"], o)
            )

    for it in writers.keys():
        writers[it].flush()
    for it in objects.keys():
        objects[it].close()

    return state


def send_usage_stats() -> None:
    try:
        version = pkg_resources.get_distribution("target-avro").version
        conn = http.client.HTTPConnection("collector.singer.io", timeout=10)
        conn.connect()
        params = {
            "e": "se",
            "aid": "singer",
            "se_ca": "target-storage",
            "se_ac": "open",
            "se_la": version,
        }
        conn.request("GET", "/i?" + urllib.parse.urlencode(params))
        conn.getresponse()
        conn.close()
    except Exception as err:
        logger.debug("Collection request failed")
        logger.debug(err)


def main() -> None:
    args = singer.utils.parse_args(REQUIRED_CONFIG_KEYS)
    config = args.config

    logger.setLevel(config.get("logging_level", logging.INFO))
    if not config.get("disable_collection", False):
        logger.info(
            "Sending version information to singer.io. "
            + "To disable sending anonymous usage data, set "
            + 'the config parameter "disable_collection" to true'
        )
        threading.Thread(target=send_usage_stats).start()

    state = persist_lines(config, io.TextIOWrapper(sys.stdin.buffer, encoding="utf-8"))

    emit_state(state)
    logger.debug("Exiting normally")


if __name__ == "__main__":
    main()



import random
import string

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import dateutil.parser


def _random_name(n):
    return "".join(random.choices(string.ascii_letters + string.digits, k=n))


def _process_json2avro(
    schema: Dict[str, Any], t: str
) -> Optional[Union[str, Dict[str, Any]]]:
    if t == "object":
        fields = json2avro(schema.get("properties"), schema.get("required", []))
        if not fields:
            return None
        return {"name": _random_name(10), "type": "record", "fields": fields}
    elif t == "array":
        if "items" not in schema:
            # ignore array without items
            return None
        # TODO: check items type, dict or list.
        if "type" not in schema["items"]:
            # ignore empty type
            return None
        next_t = schema["items"]["type"]
        types = [next_t] if isinstance(next_t, str) else next_t
        items = [_process_json2avro(schema["items"], nt) for nt in types]
        if len(items) == 0:
            return None
        if len(items) == 1:
            return {"type": "array", "items": items[0]}
        if len(items) >= 2:
            return {"type": "array", "items": items}
    elif t == "string" and schema.get("format") == "date-time":
        return {"type": "long", "logicalType": "timestamp-millis"}
    elif t == "integer":
        return "long"
    elif t == "number":
        return "double"
    return t


def json2avro(
    props: Optional[Dict[str, Any]], required: List[str]
) -> List[Dict[str, Any]]:
    if props is None:
        # ignore empty properties
        return []
    result = []
    for prop, schema in props.items():
        # TODO: check singer metadata.
        if "type" not in schema:
            # ignore empty type
            continue
        types = []
        for t in (
            [schema["type"]] if isinstance(schema["type"], str) else schema["type"]
        ):
            r = _process_json2avro(schema, t)
            if r:
                types.append(r)
        if not types or len(types) == 1 and types[0] == "null":
            # ignore empty types
            continue
        ts = ["null", *types] if prop not in required and len(types) == 1 else types
        # TODO: check avro name restriction.
        # see http://avro.apache.org/docs/current/spec.html#names
        result.append({"name": prop, "type": ts})
    return result


def _process_str2datetime(records: List[Any], schema: Dict[str, Any]) -> List[Any]:
    t = schema["type"]
    if t == "object":
        return [str2datetime(rec, schema["properties"]) for rec in records]
    elif t == "array":
        return [_process_str2datetime(recs, schema["items"]) for recs in records]
    elif t == "string" and schema.get("format") == "date-time":
        return [int(dateutil.parser.parse(rec).timestamp()) * 1000 for rec in records]
    return records


def str2datetime(record: Dict[str, Any], props: Dict[str, Any]) -> Dict[str, Any]:
    for prop, schema in props.items():
        if "type" not in schema:
            # ignore empty type
            continue
        types = [schema["type"]] if isinstance(schema["type"], str) else schema["type"]
        if "object" in types:
            if record.get(prop) is None:
                continue
            str2datetime(record[prop], schema["properties"])
        if "array" in types:
            if record.get(prop) is None:
                continue
            if "items" not in schema:
                continue
            if "type" not in schema["items"]:
                continue
            record[prop] = _process_str2datetime(record[prop], schema["items"])
        elif "string" in types and schema.get("format") == "date-time":
            if record.get(prop) is None:
                continue
            record[prop] = int(dateutil.parser.parse(record[prop]).timestamp()) * 1000
    return record



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['target_avro']

package_data = \
{'': ['*']}

install_requires = \
['fastavro>=1.3.4,<2.0.0',
 'singer-python>=5,<6',
 'smart_open>=4.2.0,<5.0.0',
 'strict_rfc3339>=0.7,<0.8']

entry_points = \
{'console_scripts': ['fmt = scripts:fmt',
                     'lint = scripts:lint',
                     'target-avro = target_avro:main',
                     'test = scripts:test']}

setup_kwargs = {
    'name': 'target-avro',
    'version': '0.1.2',
    'description': 'Singer.io target for extracting data',
    'long_description': '# target-avro\n\nThis is a [Singer](https://singer.io) target that reads JSON-formatted data\nfollowing the [Singer spec](https://github.com/singer-io/getting-started/blob/master/SPEC.md).\n\n## Features\n\n- Output Avro files for [Singer](https://singer.io) streams\n- Output to cloud storages like Google Cloud Storage and Amazon S3, etc are supported powered by [smart_open](https://github.com/RaRe-Technologies/smart_open).\n\n## Install\n\n```sh\npip install target-avro\n```\n\n## Usage\n\n```sh\n# simple\ncat <<EOF | target-avro -c sample_config.json\n{"type":"STATE","value": {}}\n{"key_properties":["id"],"schema":{"properties":{"assignee":{"properties":{},"type":["null","object"]},"created_at":{"format":"date-time","type":["null","string"]},"id":{"type":["null","integer"]},"labels":{"items":{"properties":{"id":{"type":["null","integer"]},"name":{"type":["null","string"]}},"type":"object"},"type":["null","array"]},"locked":{"type":["null","boolean"]},"pull_request":{"properties":{"url":{"type":["null","string"]}},"type":["null","object"]},"title":{"type":"string"}},"selected":true,"type":["null","object"]},"stream":"issues","type":"SCHEMA"}\n{"type": "RECORD", "stream": "issues", "record": {"created_at":"2020-11-24T23:49:24.000000Z","id":12,"labels":[{"id":238,"name":"ABCDEFGHIJKLMNOPQRSTUV"}],"locked":true,"pull_request":{"url":"https://api.github.com/repos/sample/issues/pulls/999999"},"title":"ABCDEFGHIJKLMNOPQRSTUVWXY"}, "time_extracted": "2021-03-25T12:53:51.817781Z"}\n{"type": "STATE", "value": {"bookmarks": {"singer-io/singer-python": {"issues": {"since": "2020-11-24T23:49:24.000000Z"}}}}}\nEOF\n\n# complex\ncat ./tests/data/github.jsonl | target-avro -c sample_config.json\n```\n\n## Configuration\n\nThe fields available to be specified in the config file are specified\nhere.\n\n| Field | Type | Default | Details |\n| ---- | ---- | ---- | ---- |\n| `prefix` | `["string"]`  | `N/A` | The output uri prefix. See [smart_open](https://github.com/RaRe-Technologies/smart_open) for information about valid values and credentials. |\n| `disable_collection` | `["boolean", "null"]` | `false` | Include `true` in your config to disable [Singer Usage Logging](#usage-logging). |\n| `logging_level` | `["string", "null"]`  | `"INFO"` | The level for logging. Set to `DEBUG` to get things like HTTP requests executed, JSON and Avro schemas, etc. See [Python\'s Logger Levels](https://docs.python.org/3/library/logging.html#levels) for information about valid values. |\n\n## Known Limitations\n\n- Requires a [JSON Schema](https://json-schema.org/) for every stream.\n- Only string, string with date-time format, integer, number, boolean,\n  object, and array types with or without null are supported. Arrays can\n  have any of the other types listed, including objects as types within\n  items.\n    - Example of JSON Schema types that work\n        - `[\'number\']`\n        - `[\'string\']`\n        - `[\'string\', \'null\']`\n    - Exmaple of JSON Schema types that **DO NOT** work\n        - `[\'string\', \'integer\']`\n        - `[\'integer\', \'number\']`\n        - `[\'any\']`\n        - `[\'null\']`\n- JSON Schema combinations such as `anyOf` and `oneOf` are not supported.\n- JSON Schema `$ref` is not supported.\n\n## Usage Logging\n[Singer.io](https://www.singer.io/) requires official taps and targets to collect anonymous usage data. This data is only used in aggregate to report on individual tap/targets, as well as the Singer community at-large. IP addresses are recorded to detect unique tap/targets users but not shared with third-parties.\n\nTo disable anonymous data collection set disable_collection to true in the configuration JSON file.\n\n---\n\nCopyright &copy; 2021 Kageboushi\n',
    'author': 'inamura',
    'author_email': 'inamura@kageboushi.app',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kageboushi-app/target-avro',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
